###############################################################################
#
#   Kaivo Public Software License
#
#   Copyright (c) 2001, Kaivo, Inc.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are
#   met:
#
#       o Redistributions of source code must retain the above copyright
#         notice, this list of conditions and the following disclaimer.
#
#       o Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer
#         in the documentation and/or other materials provided with the
#         distribution.
#
#       o Neither Kaivo nor the names of its contributors may be used to
#         endorse or promote products derived from this software without 
#         specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED. IN NO EVENT SHALL KAIVO OR ITS CONTRIBUTORS BE LIABLE FOR 
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
#   CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#   LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
#   USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
###############################################################################

import re
from string import strip
from DocumentTemplate.DT_Util import Eval
try: from DocumentTemplate.DT_Util import expr_globals
except: expr_globals = None
from Acquisition import Implicit
from Products.ZCatalog.Lazy import LazyMap
from BTrees.IIBTree import intersection, weightedIntersection, union, weightedUnion, difference, IISet

CQL_re = re.compile(
    ('\s*([_\-a-zA-Z0-9]+)\s*'
     '(==|>=|<=|>|<|!=|\sin\s|\snot\s+in\s|\sbetween\s)\s*'
     '(.*?(?=\sand|\sor)|.*)'
     '(?:\s+(and|or)\s+)?')
)

class ExprBlock(Implicit):
    """Query expression block"""

    index = ''      # Index name
    operator = ''   # Logical operator for value
    expr = None     # value expression
    logic = ''      # logical block operator (and | or)
    
    def exprValue(self, md={}):
        """Return the value of the criteria expression"""
        if not hasattr(self.aq_base, '_v_expr_value'):
            self._v_expr_value = self.expr.eval(md)
        return self._v_expr_value

    def exprMap(self, md={}, map={}):
        """Return the appropriate mapping for the expression
           block for use by the index machinery"""
        map[self.index] = self.exprValue(md)       
        op = self.operator
        if op in ('>=', '>'):
            map[self.index + '_usage'] = 'range:min'
        elif op in ('<=', '<'):
            map[self.index + '_usage'] = 'range:max'
        elif op == 'between':
            map[self.index + '_usage'] = 'range:min:max'

        return map

    def __call__(self, md={}):
        """Execute the query block and return the result set"""

        # Find needed objects through acquisition
        query = self.aq_parent
        indexes = query._catalog.indexes
        if not indexes.has_key(self.index):
            raise 'CatalogQueryError', 'Index "%s" not found in Catalog' % self.index
        index = indexes[self.index].__of__(self)

        if hasattr(index, '_apply_index'):
            rs, u = index._apply_index(self.exprMap(md))

            # Perform additional logic for certain operators
            op = self.operator
            if op in ('<', '>'):
                drs, u = index._apply_index({self.index: self.exprValue(md)})
                rs = difference(rs, drs)
            elif op == '!=' or op[:3] == 'not':
                all_keys = query._getAllKeys()
                rs = difference(all_keys, rs) 

            return (rs, u)
        else:
            return (None, (self.index,))


class CatalogQuery(Implicit):
    """Zope Catalog query class"""
    
    def __init__(self, catalog, expr):
        self.expr = expr
        self._catalog = catalog
        self.parse(expr)

    def parse(self, expr):
        """Parse a query string into a sequence of ExprBlocks"""
        expr_split = CQL_re.split(expr)
#        print expr_split

        if not expr_split: 
            raise 'ParseError', 'Invalid query expression string'

        more = 1
        start = 0
        logic = None
        expr_blocks = []
        while more:
            if expr_split[start]:
                raise 'ParseError', 'Parse error in query expression near: "%s"' % expr_split[start]

            block = ExprBlock()
            block = block.__of__(self)
            block.index = expr_split[start+1]
            block.operator = strip(expr_split[start+2])
            if expr_globals is not None:
                block.expr = Eval(expr_split[start+3], expr_globals) # Old Expr API
            else:
                block.expr = Eval(expr_split[start+3]) # New 2.4 API                
            block.logic = logic
            expr_blocks.append(block)
            logic = expr_split[start+4]
            start = start + 5
            more = logic or expr_split[start]

            if more and start >= len(expr_split):
                raise 'ParseError', 'Unexpected end of line in query expression'

#        for block in expr_blocks:
#            print '---'
#            print 'index: %s' % block.index
#            print 'operator: %s' % block.operator
#            print 'value: %s' % block.expr
#            print 'logic: %s' % block.logic

        self._blocks = expr_blocks
    
    def _getAllKeys(self):
        """Return a set of keys to all items in the catalog"""
        if not hasattr(self, '_v_all_keys'):
            self._v_all_keys = IISet(self._catalog.data.keys())
        return self._v_all_keys

    def _cleanUp(self):
        """Clean up after executing by purging cached data"""
        if hasattr(self, '_v_all_keys'):
            del self._v_all_keys

    def __call__(self, md={}, used=None, **kw):
        """Execute the query on the catalog"""
        
        # Combine any keywords into the namespace mapping
        if kw:
            md.update(kw)

        # Execute each query block and logically combine the results
        rs = None
        for block in self._blocks:
            r, u = block(md)

            if r is not None:
                if block.logic is None:
                    rs = r
                elif block.logic == 'and':
                    w, rs = weightedIntersection(rs, r)
                elif block.logic == 'or':
                    w, rs = weightedUnion(rs, r)

        # Create a lazy result map of the results
        if hasattr(rs, 'keys'): rs=rs.keys() 
        results = LazyMap(self._catalog.__getitem__, rs)
        
        self._cleanUp()
        return results
